# Copyright (c) 2023 TheraPanacea. All rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited
# Proprietary and confidential
"""A client for the ART platform. Wraps HTTP calls."""

import logging
from typing import Any

import requests


class ARTClient:
    """A client for the ART platform."""

    def __init__(
        self,
        base_url: str = 'http://radiomics.thera/dev',
        *,
        username: str,
        password: str,
        auth_route: str = 'auth',
        routes: dict[str, str] = None,
        router_disabled=False,
    ):
        self._base_url = base_url
        self._auth_route = auth_route
        self._username = username
        self._password = password
        self._router_disabled = router_disabled

        routes = {} if routes is None else routes
        default_routes = {
            'data': 'datamanager',
            'sdk': 'sdk',
            'scheduler': 'scheduler',
        }
        self._routes = {**default_routes, **routes}
        self._token = None

    def service(self, name: str) -> str:
        return self._routes.get(name, '')

    def request(
        self,
        fn,
        service: str,
        route: str,
        *args,
        max_retries: int = 3,
        with_auth_headers: bool = True,
        **kwargs
    ) -> requests.Response:
        """A generic function to call a HTTP verb on a URL, ensuring login."""
        route = route.lstrip('/')
        if self._router_disabled:
            url = f'{self._base_url}:{service}/{route}'
        else:
            service = service.rstrip('/')
            url = f'{self._base_url}/{service}/{route}'
        headers = {'Authorization': f'Bearer {self._token}'}
        headers = headers if with_auth_headers else None
        resp = fn(url, *args, headers=headers, **kwargs)
        if resp.status_code == 401 and max_retries > 0:
            max_retries -= 1
            self.login()
            return self.request(
                fn,
                service,
                route,
                *args,
                max_retries=max_retries,
                with_auth_headers=with_auth_headers,
                **kwargs
            )
        return resp

    def login(self) -> str | None:
        """Get a fresh auth token."""
        credentials = {'username': self._username, 'password': self._password}
        resp = self.post(
            self._auth_route,
            'token/login',
            data=credentials,
            timeout=5,
            max_retries=0
        )
        if resp.status_code == 200:
            self._token = resp.json().get('access_token')
            return self._token
        else:
            logging.error(
                'User %s cannot login to %s', self._username, self._base_url
            )
            return None

    def post(self, service: str, route: str, *args, **kwargs):
        return self.request(requests.post, service, route, *args, **kwargs)

    def get(self, service: str, route: str, *args, **kwargs):
        return self.request(requests.get, service, route, *args, **kwargs)

    def put(self, service, route, *args, **kwargs):
        return self.request(requests.put, service, route, *args, **kwargs)
