# Copyright (c) 2023 TheraPanacea. All rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited
# Proprietary and confidential
"""A client for the ART platform. Wraps HTTP calls."""

import glob
import http.client
import io
import json
import logging
import os
from typing import Any, BinaryIO
import uuid

import nibabel as nib
import numpy as np
import pandas as pd
import requests

from artist import client as client_lib
from artist import tqdm


def push_data(
    client: client_lib.ARTClient,
    paths: str | list[str],
    data_format_name: str,
    meta_data: dict[str, Any] | str | None = None,
    minio_host: str | None = None,
    **tqdm_kwargs
) -> list[str]:
    """Pushes data to the art platform, with optional metadata."""
    allowed_formats = ('NIFTI', 'DICOM', 'DICOM_ZIP')
    if data_format_name not in allowed_formats:
        logging.error(
            'Cannot load %s. Allowed formats are %s', data_format_name,
            allowed_formats
        )
        return

    if isinstance(paths, str) and os.path.isdir(paths):
        paths = glob.glob(os.path.join(paths, '*.dcm'))

    if isinstance(meta_data, str) and os.path.exists(meta_data):
        df = pd.read_csv(meta_data).fillna('')
        meta_data = df.iloc[0].to_dict()

    data_root = client.service('data')
    result = []
    paths = [paths] if isinstance(paths, str) else paths
    for path in tqdm.tqdm(paths, **tqdm_kwargs):
        r = client.post(
            data_root, 'datas', json={'data_format_name': data_format_name}
        )
        if r.status_code != http.client.OK:
            logging.error('Error %s asking for upload url', r.status_code)

        resp = r.json()
        url = resp['url']
        with open(path, 'rb') as fp:
            headers = {'Content-type': 'application/octet-stream'}
            if minio_host is not None:
                headers['HOST'] = minio_host
            r = requests.put(url, data=fp, headers=headers, timeout=600)
        if r.status_code != http.client.OK:
            logging.error('Error %s uploading files.', r.status_code)

        data_uid = resp['data_uid']
        r = client.put(data_root, f'datas/{data_uid}', json={'stored': True})
        if r.status_code != http.client.OK:
            logging.error('Error %s notifying upload', r.status_code)
        else:
            data_uid = r.json()['data_uid']

        if meta_data is not None:
            r = client.put(
                data_root,
                f'/datas/{data_uid}/meta_data',
                json={'meta_data': meta_data}
            )
            if r.status_code != http.client.OK:
                logging.error('Error %s adding metadata', r.status_code)

        if data_uid not in result:
            result.append(data_uid)
    return result


def create_cohort(client: client_lib.ARTClient, name: str) -> str | None:
    """Creates a cohort with a name. Returns the new cohort_id."""
    data_root = client.service('data')
    r = client.post(data_root, 'cohorts', json={'name': name})
    if r.status_code != http.client.OK:
        logging.error('Error creating cohort (%s): %s', name, r.content)
        return None

    resp = r.json()
    return resp['cohort_uid']


def add_cohort_data(
    client: client_lib.ARTClient, cohort_uid: str, data_uids: list[str]
) -> bool:
    """Adds data to a cohort."""
    data_root = client.service('data')
    r = client.put(
        data_root, f'cohorts/{cohort_uid}', json={'data_uids': data_uids}
    )
    if r.status_code != http.client.OK:
        logging.error(
            'Error adding data to cohort (%s): %s', cohort_uid, r.content
        )
    return r.status_code == http.client.OK


def get_cohort_data(client: client_lib.ARTClient, cohort_uid: str) -> list[str]:
    """Returns the list of uids of data from a cohort."""
    data_root = client.service('data')
    r = client.get(data_root, f'cohorts/{cohort_uid}')
    if r.status_code != http.client.OK:
        logging.error(
            'Error getting cohort data (%s): %s', cohort_uid, r.content
        )
        return []

    resp = r.json()
    return resp['data_uids']


def get_header(client: client_lib.ARTClient, data_uid: str) -> list[str]:
    """Returns the header of a given data."""
    data_root = client.service('data')
    r = client.get(data_root, f'datas/headers/{data_uid}')
    if r.status_code != http.client.OK:
        logging.error('Error getting header (%s): %s', data_uid, r.content)
        return []

    resp = r.json()
    return resp


def download(url: str, fp: BinaryIO):
    """Downloads the content of a file at a URL on disk."""
    resp = requests.get(url, timeout=300)
    if resp.status_code != http.client.OK:
        logging.error('Cannot download %s', url)
        raise RuntimeError(
            f'Cannot download {url}, got error {resp.status_code}: {resp.content}'
        )

    fp.write(resp.content)


def fetch_data(
    client: client_lib.ARTClient,
    data_uid: str,
    output: str,
    data_format: str = '',
):
    """Downloads the data_uid and saves it locally."""
    data_root = client.service('data')
    resp = client.get(
        data_root,
        f'datas/{data_uid}',
        params={'data_format_name': data_format}
    )
    if resp.status_code != http.client.OK:
        logging.error('Cannot fetch %s', data_uid)
        return

    urls = resp.json().get('urls', None)
    if urls is None:
        logging.error('No url associated with %s', data_uid)
        return

    if len(urls) == 1:
        os.makedirs(os.path.dirname(output), exist_ok=True)
        with open(output, 'wb') as fp:
            download(urls[0], fp)
    else:
        os.makedirs(output, exist_ok=True)
        for url in tqdm.tqdm(urls):
            with open(os.path.join(output, f'{uuid.uuid4()}.dcm'), 'wb') as fp:
                download(url, fp)


def search(client: client_lib.ARTClient, query: dict[str, Any]) -> list[str]:
    """Search for data_uids matching the input query."""
    data_root = client.service('data')
    req = {'query': json.dumps(query), 'fields': ['field:data_uid']}
    r = client.get(data_root, 'datas/query', params=req)
    if r.status_code != http.client.OK:
        logging.error(
            'Error when querying DataManager (query: %s)', req['query']
        )
        return []
    return [r['data_uid'] for r in r.json()['results']]


def load_image(stream: BinaryIO, data_format: str):
    """Loads the content of a stream as nifit."""
    stream.seek(0)
    if data_format == 'NIFTI':
        file_holder = nib.FileHolder(fileobj=stream)
        nib_img = nib.Nifti1Image.from_file_map(
            {
                'header': file_holder,
                'image': file_holder
            }
        )
        return nib_img
    if data_format in ('NUMPY', 'NUMPY_ZIP'):
        data = np.load(stream, allow_pickle=True)
        data = data[()] if () in data else data
        return dict(data)

    raise ValueError(f'Format {data_format} is not handled by `load_image`.')


def download_image(url: str, data_format: str):
    stream = io.BytesIO()
    download(url, stream)
    stream.seek(0)
    return load_image(stream, data_format=data_format)


def get_jobs(client: client_lib.ARTClient, limit: int = 10) -> pd.DataFrame:
    """Retrieves the last jobs that have been updated."""
    resp = client.get(client.service('scheduler'), f'jobs?limit={limit}')
    if resp.status_code == 200:
        return pd.DataFrame(resp.json())
